#ifndef MP4V2_PROJECT_H
#define MP4V2_PROJECT_H

/*****************************************************************************/

#define MP4V2_PROJECT_name            "MP4v2"
#define MP4V2_PROJECT_name_lower      "mp4v2"
#define MP4V2_PROJECT_name_upper      "MP4V2"
#define MP4V2_PROJECT_name_formal     "MP4v2 -r"
#define MP4V2_PROJECT_url_website     "http://code.google.com/p/mp4v2"
#define MP4V2_PROJECT_url_downloads   "http://code.google.com/p/mp4v2/downloads/list"
#define MP4V2_PROJECT_url_discussion  "http://groups.google.com/group/mp4v2"
#define MP4V2_PROJECT_irc             "irc://irc.freenode.net/handbrake"
#define MP4V2_PROJECT_bugreport       "<kidjan@gmail.com>"
#define MP4V2_PROJECT_version         "-r"
#define MP4V2_PROJECT_version_hex     0x00020100
#define MP4V2_PROJECT_version_major   2
#define MP4V2_PROJECT_version_minor   1
#define MP4V2_PROJECT_version_point   0
#define MP4V2_PROJECT_repo_url        "svn://nowhere.com/project/unknown"
#define MP4V2_PROJECT_repo_branch     "unknown"
#define MP4V2_PROJECT_repo_root       "svn://nowhere.com/project"
#define MP4V2_PROJECT_repo_uuid       "00000000-0000-0000-0000-000000000000"
#define MP4V2_PROJECT_repo_rev        0
#define MP4V2_PROJECT_repo_date       "unknown"
#define MP4V2_PROJECT_repo_type       "developer"
#define MP4V2_PROJECT_build           "Fri Feb 24 09:32:44 CET 2017"

/*****************************************************************************/

#endif /* MP4V2_PROJECT_H */