/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

/* 
 * File:   RTPPacketSched.cpp
 * Author: Sergio
 * 
 * Created on 3 de febrero de 2017, 11:53
 */

#include "rtp.h"

