#ifndef MP4V2_PLATFORM_BASE_H
#define MP4V2_PLATFORM_BASE_H

///////////////////////////////////////////////////////////////////////////////

#include <fstream>
#include <iomanip>
#include <iostream>
#include <limits>
#include <list>
#include <locale>
#include <map>
#include <set>
#include <sstream>
#include <string>
#include <vector>

#include <cassert>
#include <cctype>
#include <cerrno>
#include <climits>
#include <cmath>
#include <cstdarg>
#include <cstddef>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <cwchar>
#include <cwctype>

///////////////////////////////////////////////////////////////////////////////

#endif // MP4V2_PLATFORM_BASE_H
