#include "EventSource.h"

EvenSource::EvenSource() {}
EvenSource::EvenSource(const char* str) {}
EvenSource::EvenSource(const std::wstring &str) {}
EvenSource::~EvenSource() {}
void EvenSource::SendEvent(const char* type,const char* msg,...) {}
