#ifndef MP4V2_PLATFORM_PROCESS_PROCESS_H
#define MP4V2_PLATFORM_PROCESS_PROCESS_H

/// @namespace mp4v2::platform::process (private) Process.
/// <b>WARNING: THIS IS A PRIVATE NAMESPACE. NOT FOR PUBLIC CONSUMPTION.</b>
namespace mp4v2 { namespace platform { namespace process {

///////////////////////////////////////////////////////////////////////////////

MP4V2_EXPORT int32_t getpid();

///////////////////////////////////////////////////////////////////////////////

}}} // namespace mp4v2::platform::process

#endif // MP4V2_PLATFORM_PROCESS_PROCESS_H
