#ifndef MP4V2_PROJECT_H
#define MP4V2_PROJECT_H

/*****************************************************************************/

#define MP4V2_PROJECT_name            ""
#define MP4V2_PROJECT_name_lower      ""
#define MP4V2_PROJECT_name_upper      ""
#define MP4V2_PROJECT_name_formal     ""
#define MP4V2_PROJECT_url_website     ""
#define MP4V2_PROJECT_url_downloads   ""
#define MP4V2_PROJECT_url_discussion  ""
#define MP4V2_PROJECT_irc             ""
#define MP4V2_PROJECT_bugreport       ""
#define MP4V2_PROJECT_version         ""
#define MP4V2_PROJECT_version_hex     
#define MP4V2_PROJECT_version_major   
#define MP4V2_PROJECT_version_minor   
#define MP4V2_PROJECT_version_point   
#define MP4V2_PROJECT_repo_url        ""
#define MP4V2_PROJECT_repo_branch     ""
#define MP4V2_PROJECT_repo_root       ""
#define MP4V2_PROJECT_repo_uuid       ""
#define MP4V2_PROJECT_repo_rev        0
#define MP4V2_PROJECT_repo_date       ""
#define MP4V2_PROJECT_repo_type       ""
#define MP4V2_PROJECT_build           ""

/*****************************************************************************/

#endif /* MP4V2_PROJECT_H */
